/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <ctype.h>
#include <stdlib.h>

#include "alloc.h"
#include "matrix.h"
#include "affine.h"
#include "list.h"
#include "rational.h"
#include "calc.h"
#include "output.h"


int G2_entries[] = {
   2, -1, 
  -3,  2
};

int F4_entries[] = { 
   2, -1,  0,  0,
  -1,  2, -2,  0,
   0, -1,  2, -1,
   0,  0, -1,  2
};

int E8_entries[] = {
   2,  0, -1,  0,  0,  0,  0,  0,
   0,  2,  0, -1,  0,  0,  0,  0,
  -1,  0,  2, -1,  0,  0,  0,  0,
   0, -1, -1,  2, -1,  0,  0,  0,
   0,  0,  0, -1,  2, -1,  0,  0,
   0,  0,  0,  0, -1,  2, -1,  0,
   0,  0,  0,  0,  0, -1,  2, -1,
   0,  0,  0,  0,  0,  0, -1,  2
};

matrix *mk_cartan(char *s)
{
  char typ = toupper(*s);
  int n = atoi(s + 1);
  matrix *c = NULL;
  int i, j;
  
  if (n <= 0)
    return NULL;
  
  switch (typ)
    {
    case 'E':
      if (6 <= n && n <= 8)
	{
	  c = m_new(n);
	  for (i = 0; i < n; i++)
	    for (j = 0; j < n; j++)
	      m_entry(c, i, j) = E8_entries[8 * i + j];
	}
      break;
      
    case 'F':
      if (n == 4)
	{
	  c = m_new(4);
	  for (i = 0; i < 4; i++)
	    for (j = 0; j < 4; j++)
	      m_entry(c, i, j) = F4_entries[4 * i + j];
	}
      break;
      
    case 'G':
      if (n == 2)
	{
	  c = m_new(2);
	  for (i = 0; i < 2; i++)
	    for (j = 0; j < 2; j++)
	      m_entry(c, i, j) = G2_entries[2 * i + j];
	}
      break;
      
    case 'D':
      if (n < 4) break;
    case 'B':
    case 'C':
      if (n < 2) break;
    case 'A':
      c = m_new_id(n);
      m_scale(c, 2, c);
      for (i = 1; i < n; i++)
	{
	  m_entry(c, i - 1, i) = -1;
	  m_entry(c, i, i - 1) = -1;
	}
      if (n > 1)
	{
	  if (typ == 'B')
	    m_entry(c, n - 2, n - 1) = -2;
	  else if (typ == 'C')
	    m_entry(c, n - 1, n - 2) = -2;
	  else if (typ == 'D')
	    {
	      m_entry(c, n - 2, n - 1) = m_entry(c, n - 1, n - 2) = 0;
	      if (n > 2)
		m_entry(c, n - 3, n - 1) = m_entry(c, n - 1, n - 3) = -1;
	    }
	}
      break;
    }
  
  return c;
}


int find_pos_column(matrix *w)
{
  int n = m_dim(w);
  int i, j;
  for (j = 0; j < n; j++)
    {
      for (i = 0; i < n && m_entry(w, i, j) == 0; i++)
	;
      if (i < n && m_entry(w, i, j) > 0)
	return j + 1;
    }
  return 0;
}

static void fill_simp_len(int *simp_len, int i, matrix *cartan)
{
  int j;
  for (j = 0; j < m_dim(cartan); j++)
    if (simp_len[j] == 0 && m_entry(cartan, i, j) != 0)
      {
	simp_len[j] =
	  (simp_len[i] * m_entry(cartan, j, i)) / m_entry(cartan, i, j);
	fill_simp_len(simp_len, j, cartan);
      }
}

affine *init_affine(char *typ)
{
  int n, p, i, j, det_helm;
  char c, *ppos;
  matrix *cartan;
  affine *afn;
  matrix **refmat;
  matrix *w0, *w1;
  vector *v0, *v1;
  int ht, maxht;
  list *plusroots, *proots, *rlength;
  int *simp_len;
  matrix *rdot, *wdot;
  int wdot_fac, gcd;
  
  p = 0;
  ppos = typ + 1;
  while (isdigit(*ppos))
    ppos++;
  if (*ppos == '-')
    {
      ppos++;
      p = atoi(ppos);
      if (p <= 0)
	return NULL;
      while (isdigit(*ppos))
	ppos++;
    }
  if (*ppos != '\0')
    return NULL;
  
  cartan = mk_cartan(typ);
  if (cartan == NULL)
    return NULL;
  n = m_dim(cartan);
  
  afn = (affine *) weyl_malloc(sizeof(affine));
  afn->cartan = cartan;
  afn->root_system_type = toupper(*typ);
  
  afn->tcartan = m_new(n);
  m_transpose(afn->tcartan, cartan);
  
  afn->acartan = m_new(n);
  m_adjoint(afn->acartan, cartan, & afn->dcartan);
  
  afn->atcartan = m_new(n);
  m_transpose(afn->atcartan, afn->acartan);
  
  simp_len = (int *) weyl_calloc(n, sizeof(int));
  c = toupper(*typ);
  simp_len[0] = (c == 'B' || c == 'F') ? 2 : 1;
  fill_simp_len(simp_len, 0, cartan);
  afn->simple_length = simp_len;
  
  rdot = m_new(n);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      m_entry(rdot, i, j) = simp_len[j] * m_entry(cartan, i, j);
  afn->root_dot = rdot;
  
  wdot = m_new(n);
  wdot_fac = 2 * afn->dcartan * afn->dcartan;
  w1 = m_new(n);
  m_mult(w1, rdot, afn->atcartan);
  m_mult(wdot, afn->acartan, w1);
  gcd = m_entry(wdot, 0, 0);
  for (i = 0; i < n; i++)
    for (j = i; j < n; j++)
      gcd = i_gcd(gcd, m_entry(wdot, i, j));
  gcd = i_gcd(gcd, wdot_fac);
  wdot_fac /= gcd;
  m_div(wdot, wdot, gcd);
  afn->wdot_factor = wdot_fac;
  afn->weight_dot = wdot;
  
  afn->cartan_rows = (vector **) weyl_malloc(n * sizeof(vector *));
  afn->cartan_columns = (vector **) weyl_malloc(n * sizeof(vector *));
  for (i = 0; i < n; i++)
    {
      afn->cartan_rows[i] = v_new(n);
      m_get_row(afn->cartan_rows[i], cartan, i);
      afn->cartan_columns[i] = v_new(n);
      m_get_column(afn->cartan_columns[i], cartan, i);
    }
  
  refmat = (matrix **) weyl_malloc(n * sizeof(matrix *));
  for (i = 0; i < n; i++)
    {
      refmat[i] = m_new_id(n);
      for (j = 0; j < n; j++)
	m_entry(refmat[i], i, j) -= m_entry(afn->tcartan, i, j);
    }
  
  afn->highest_root = v_new(n);
  afn->highest_elem = m_new(n);
  w0 = m_new_id(n);
  v0 = v_new(n);
  proots = l_new();
  rlength = l_new();
  maxht = 0;
  while ((i = find_pos_column(w0)) != 0)
    {
      int length;
      vector *new = v_new(n);
      m_get_column(new, w0, i - 1);
      
      l_appendto(proots, new);
      m_multvec(v0, rdot, new);
      length = v_dot(new, v0) / 2;
      l_appendint(rlength, length);
      
      if (length == 1 && (ht = v_sum_of_entries(new)) > maxht)
	{
	  v_copy(afn->highest_root, new);
	  m_copy(afn->highest_elem, w0);
	  afn->highest_simp = i;
	  maxht = ht;
	}
      m_mult(w1, w0, refmat[i - 1]);
      {
	matrix *tmp = w0;
	w0 = w1;
	w1 = tmp;
      }
    }
  m_free(w0);
  for (i = 0; i < n; i++)
    m_free(refmat[i]);
  weyl_free(refmat);
  
  plusroots = afn->positive_roots = l_new();
  afn->root_length = l_new();
  for (i = 0; i < l_length(proots); i++)
    if (v_sum_of_entries(l_element(proots, i)) != 1)
      {
	l_appendto(plusroots, l_element(proots, i));
	l_appendint(afn->root_length, l_intelem(rlength, i));
      }
  for (i = 0; i < l_length(proots); i++)
    if (v_sum_of_entries(l_element(proots, i)) == 1)
      {
	l_appendto(plusroots, l_element(proots, i));
	l_appendint(afn->root_length, l_intelem(rlength, i));
      }
  l_free(rlength);
  
  afn->weight_roots = proots;
  l_reset(proots);
  for (i = 0; i < l_length(plusroots); i++)
    {
      vector *wr = v_new(n);
      m_multvec(wr, afn->tcartan, l_element(plusroots, i));
      l_appendto(afn->weight_roots, wr);
    }
  
  afn->high_weight = v_new(n);
  m_multvec(afn->high_weight, afn->tcartan, afn->highest_root);
  
  v1 = v_new(n);
  m_get_row(v1, afn->tcartan, afn->highest_simp - 1);
  m_adjoint(w1, afn->highest_elem, & det_helm);
  m_vecmult(v0, v1, w1);
  m_vecmult(v1, v0, afn->atcartan);
  v_div(v1, v1, afn->dcartan * det_helm);
  afn->high_pair = v1;
  m_free(w1);
  
  v_set_ones(v0);
  afn->rho = v0;
  
  afn->permutations = l_new();
  {
    vector *prm = v_new(n);
    for (i = 0; i < n; i++)
      v_entry(prm, i) = i + 1;
    l_appendto(afn->permutations, prm);
  }
  switch (afn->root_system_type)
    {
    case 'A':
      if (n > 1)
	{
	  vector *prm = v_new(n);
	  for (i = 0; i < n; i++)
	    v_entry(prm, i) = n - i;
	  l_appendto(afn->permutations, prm);
	}
      break;
    case 'D':
      if (n == 4)
	{
	  for (i = 0; i < 3; i++)
	    for (j = 0; j < 2; j++)
	      if (i != 0 || j != 0)
		{
		  vector *prm = v_new(4);
		  int x = 2 + i;
		  v_entry(prm, 0) = (x == 2) ? 1 : x;
		  x = 2 + (i + 1 + j) % 3;
		  v_entry(prm, 2) = (x == 2) ? 1 : x;
		  x = 2 + (i + 2 - j) % 3;
		  v_entry(prm, 3) = (x == 2) ? 1 : x;
		  v_entry(prm, 1) = 2;
		  l_appendto(afn->permutations, prm);
		}
	}
      else
	{
	  vector *prm = v_new(n);
	  for (i = 0; i < n - 2; i++)
	    v_entry(prm, i) = i + 1;
	  v_entry(prm, n - 2) = n;
	  v_entry(prm, n - 1) = n - 1;
	  l_appendto(afn->permutations, prm);
	}
      break;
    case 'E':
      if (n == 6)
	{
	  vector *prm = v_new(6);
	  v_entry(prm, 1 -1) = 6;
	  v_entry(prm, 2 -1) = 2;
	  v_entry(prm, 3 -1) = 5;
	  v_entry(prm, 4 -1) = 4;
	  v_entry(prm, 5 -1) = 3;
	  v_entry(prm, 6 -1) = 1;
	  l_appendto(afn->permutations, prm);
	}
      break;
    }
  
  afn->h = 2 * l_length(plusroots) / n;
  afn->p = p ? p : afn->h;	/*  THIS IS NOT A PRIME  */
  
  {
    vector *a = v_new(n);
    v_copy(a, afn->rho);
    v_scale(a, afn->p, a);
    
    afn->top_alcove = l_new();
    weight_to_exp(afn->top_alcove, a, NULL, afn);
    
    v_copy(a, afn->rho);
    v_scale(a, -1, a);
    afn->longest_element = l_new();
    alcove_to_exp(afn->longest_element, a, afn);
    
    v_free(a);
  }
  
  return afn;
}

void afn_free(affine *afn)
{
  int dim = m_dim(afn->cartan);
  int i;
  
  m_free(afn->cartan);
  m_free(afn->tcartan);
  m_free(afn->acartan);
  m_free(afn->atcartan);
  
  weyl_free(afn->simple_length);
  m_free(afn->root_dot);
  m_free(afn->weight_dot);
  
  l_free(afn->top_alcove);
  l_free(afn->longest_element);
  free_veclist(afn->positive_roots);
  free_veclist(afn->weight_roots);
  l_free(afn->root_length);
  free_veclist(afn->permutations);
  
  v_free(afn->highest_root);
  v_free(afn->highest_elem);
  
  v_free(afn->high_pair);
  v_free(afn->high_weight);
  v_free(afn->rho);
  
  for (i = 0; i < dim; i++)
    {
      v_free(afn->cartan_rows[i]);
      v_free(afn->cartan_columns[i]);
    }
  weyl_free(afn->cartan_rows);
  weyl_free(afn->cartan_columns);
  
  weyl_free(afn);
}
