/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#include "machine.h"
#include "alloc.h"
#include "matrix.h"
#include "affine.h"
#include "calc.h"
#include "tree.h"
#include "list.h"
#include "hashtab.h"
#include "output.h"

void apply_to_alcove(vector *a, list *s, affine *afn)
{
  int l, i;
  for (l = l_length(s) - 1; l >= 0; l--)
    {
      i = (int) l_element(s, l);
      ref_weight(a, i, afn);
    }
}

void calc_alcove(vector *a, list *s, affine *afn)
{
  int l, i;
  v_set_ones(a);
  for (l = l_length(s) - 1; l >= 0; l--)
    {
      i = (int) l_element(s, l);
      ref_weight(a, i, afn);
    }
}

int weight_to_exp(list *exp, vector *weight, vector *res, affine *afn)
{
  int n = v_length(weight);
  vector *a = (res == NULL) ? v_new(n) : res;
  int s;
  
  v_copy(a, weight);
  l_reset(exp);
  
  while (1)
    {
      s = n;
      while (s > 0 && v_entry(a, s - 1) >= 0)
	s--;
      if (s == 0)
	{
	  int pair = v_dot(a, afn->high_pair);
	  if (pair <= afn->p)
	    break;
	  
	  v_add_multiple(a, (afn)->p - pair, (afn)->high_weight);
	  l_appendint(exp, s);
	}
      else
	{
	  v_add_multiple(a, - v_entry(a, s - 1), afn->cartan_rows[s - 1]);
	  l_appendint(exp, s);
	}
    }
  for (s = 0; s < n && v_entry(a, s) > 0; s++)
    ;
  if (s == n && v_dot(a, afn->high_weight) == afn->p)
    s--;
  if (res == NULL)
    v_free(a);
  return s != n;
}

void alcove_to_exp(list *exp, vector *alc, affine *afn)
{
  int n = v_length(alc);
  vector *a = v_new_copy(alc);
  
  l_reset(exp);
  while (v_compare(a, afn->rho) != 0)
    {
      int i;
      for (i = n; i > 0 && v_entry(a, i - 1) >= 0; i--)
	;
      ref_weight(a, i, afn);
      l_appendto(exp, (void *) i);
    }
  v_free(a);
}

void reduced_exp(list *exp, list *s, affine *afn)
{
  vector *a = v_new(m_dim(afn->cartan));
  calc_alcove(a, s, afn);
  alcove_to_exp(exp, a, afn);
  v_free(a);
}


static int chk_element(vector *v, int filter, affine *afn)
{
  int p = afn->p;
  int i;
  
  switch (filter)
    {
    case TB_ALL:
      return 1;
    
    case TB_DOMINANT:
      for (i = 0; i < v_length(v); i++)
	if (v_entry(v, i) <= 0)
	  return 0;
      return 1;
      
    case TB_RESTRICTED:
      for (i = 0; i < v_length(v); i++)
	{
	  int x = v_entry(v, i);
	  if (x <= 0 || x > p)
	    return 0;
	}
      return 1;
      
    default:
      return 0;
    }
}

node *ins_sub_exps(treebase *tb, list *exp, int filter)
{
  affine *afn = tb_affine(tb);
  int dim = m_dim(afn->cartan);
  vector *a, *aa;
  void **nodep;
  node *node, *blw;
  list *nexp;
  int l, k, i;
  
  a = v_new(dim);
  calc_alcove(a, exp, afn);
  nodep = hash_mkfind(tb_nodeset(tb), a);
  
  if (*nodep != NULL)
    {
      v_free(a);
      return *nodep;
    }
  
  node = n_new(l_length(exp), a);
  *nodep = node;
  n_number(node) = tb_node_count(tb, l_length(exp))++;
  n_next(node) = tb_node_list(tb, l_length(exp));
  tb_node_list(tb, l_length(exp)) = node;
  
  aa = v_new(dim);
  a = v_new_copy(afn->rho);
  nexp = l_new();
  
  for (l = l_length(exp) - 1; l >= 0; l--)
    {
      v_copy(aa, a);
      for (k = l - 1; k >= 0; k--)
	{
	  i = l_intelem(exp, k);
	  ref_weight(aa, i, afn);
	}
      if (chk_element(aa, filter, afn))
	{
	  alcove_to_exp(nexp, aa, afn);
	  if (l_length(nexp) == l_length(exp) - 1)
	    {
	      blw = ins_sub_exps(tb, nexp, filter);
	      n_insbelow(node, blw);
	    }
	}
      i = l_intelem(exp, l);
      ref_weight(a, i, afn);
    }
  v_free(a);
  v_free(aa);
  l_free(nexp);
  
  return node;
}

void tb_add_element(treebase *tb, list *s, int filter)
{
  affine *afn = tb_affine(tb);
  vector *a = v_new(afn_dim(afn));
  
  calc_alcove(a, s, afn);
  if (chk_element(a, filter, afn))
    {
      list *exp = l_new();
      alcove_to_exp(exp, a, afn);
      tb_set_max_length(tb, l_length(exp));
      ins_sub_exps(tb, exp, filter);
      l_free(exp);
    }
  
  v_free(a);
}

treebase *build_tree(list *s, int filter, affine *afn)
{
  treebase *tb = tb_new(afn);
  tb_add_element(tb, s, filter);
  return tb;
}

#define set_row_bit(row,bit)	\
	((row)[(bit) / WORD_BIT]  |=  1 << ((bit) % WORD_BIT))

void or_mat_row(node *nd, int l1, unsigned *row, treebase *tb)
{
  int l2 = n_length(nd);
  int i;
  
  if (l2 - l1 == 1)
    {
      for (i = 0; i < n_nbelow(nd); i++)
	{
	  int bit = n_number(n_below(nd, i));
	  set_row_bit(row, bit);
	}
    }
  else
    {
      for (i = 0; i < n_nbelow(nd); i++)
	{
	  node *bel = n_below(nd, i);
	  unsigned *mat = tb_bit_matrix(tb, l1, l2 - 1);
	  int j;
	  int rowlen = tb_row_length(tb, l1);
	  int bel_num = n_number(bel);
	  unsigned *bel_row = mat + bel_num * rowlen;
	  for (j = 0; j < rowlen; j++)
	    row[j] |= bel_row[j];
	}
    }
}

void calc_node_bits(node *nd, treebase *tb)
{
  int l2 = n_length(nd);
  int l1;
  int num = n_number(nd);
  
  if (n_color(nd) != 0)
    return;
  n_color(nd) = 1;
  
  for (l1 = 1; l1 < l2; l1++)
    {
      unsigned *mat = tb_bit_matrix(tb, l1, l2);
      if (mat != NULL)
	or_mat_row(nd, l1, mat + tb_row_length(tb, l1) * num, tb);
    }
}

void calc_bit_matrices(treebase *tb)
{
  int max_length = tb_max_length(tb);
  int l1, l2;
  node *nd;
  
  unsigned **bit_matrices = (unsigned **)
    weyl_calloc(1, (max_length - 1) * (max_length - 1) * sizeof(unsigned *));
  tb->bit_matrices = bit_matrices;
  
  for (l1 = 1; l1 < max_length; l1++)
    for (l2 = l1 + 1; l2 <= max_length; l2++)
      tb_bit_matrix(tb, l1, l2) = (unsigned *) 
	weyl_calloc(1, tb_row_length(tb, l1) * tb_node_count(tb, l2) 
		    * sizeof(unsigned));
  
  tb_reset_node_colors(tb);
  
  for (l1 = 2; l1 <= max_length; l1++)
    for (nd = tb_node_list(tb, l1); nd != NULL; nd = n_next(nd))
      calc_node_bits(nd, tb);
}

int node_lesseq(node *nd1, node *nd2, treebase *tb)
{
  int l1, l2;
  unsigned *mat;
  int num1;
  
  if (nd1 == nd2)
    return 1;
  l1 = n_length(nd1);
  l2 = n_length(nd2);
  if (l1 == 0)
    return 1;
  else if (l1 >= l2)
    return 0;
  
  mat = tb_bit_matrix(tb, l1, l2);
  num1 = n_number(nd1);
  return (mat[tb_row_length(tb, l1) * n_number(nd2) + num1 / WORD_BIT]
	  & (1 << (num1 % WORD_BIT)));
}

int bruhat_lesseq(list *exp1, list *exp2, treebase *tb)
{
  node *n1, *n2;
  n1 = tb_find(tb, exp1);
  if (n1 == NULL)
    return 0;
  n2 = tb_find(tb, exp2);
  if (n2 == NULL)
    return 0;
  return node_lesseq(n1, n2, tb);
}
