/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "alloc.h"
#include "output.h"
#include "affine.h"
#include "matrix.h"
#include "weight.h"

int *build_action_tab(affine *afn)
{
  list *plus_roots = afn->positive_roots;
  int n = m_dim(afn->cartan);
  weight_set *rs, *nrs; 
  int *action_tab;
  vector *root = v_new(n);
  int i, j, l, dot;
  
  rs = ws_new(l_element(plus_roots, 0));
  ws_mult(rs) = 0;
  for (i = 1; i < l_length(plus_roots); i++)
    {
      weight_set *nrs = ws_insert(rs, l_element(plus_roots, i));
      ws_mult(nrs) = i;
    }
  
  action_tab = weyl_malloc(l_length(plus_roots) * n * sizeof(int));
  for (i = 0; i < n; i++)
    for (j = 0; j < l_length(plus_roots); j++)
      {
	v_copy(root, (vector *) l_element(plus_roots, j));
	dot = v_dot(root, afn->cartan_columns[i]);
	v_entry(root, i) -= dot;
	for (l = n; l > 0 && v_entry(root, l - 1) >= 0; l--)
	  ;
	if (l != 0)
	  v_scale(root, -1, root);
	nrs = ws_find(rs, root);
	action_tab[i + n * j] = ws_mult(nrs);
      }
  
  v_free(root);
  ws_free(rs);
  
  return action_tab;
}

static affine *afn;
static int *action_tab;
static vector *high_weight;
static weight_set *hw_set;
static int nplus;
static int *father_arr, *size_arr, *dom_arr;

static void reset_arrays()
{
  int i;
  memset(size_arr, 0, nplus * sizeof(int));
  memset(dom_arr, 0, nplus * sizeof(int));
  for (i = 0; i < nplus; i++)
    father_arr[i] = -1;
}

static int find_father(int r)
{
  int father;
  if (father_arr[r] == -1)
    return r;
  father = find_father(father_arr[r]);
  father_arr[r] = father;
  return father;
}

static void join_roots(int r1, int r2)
{
  r1 = find_father(r1);
  r2 = find_father(r2);
  if (r1 == r2)
    return;
  if (size_arr[r2] > size_arr[r1])
    {
      int tmp = r1;
      r1 = r2;
      r2 = tmp;
    }
  size_arr[r1] += 1 + size_arr[r2];
  father_arr[r2] = r1;
}

static int find_neg_index(vector *w)
{
  int i;
  for (i = l_length(w) - 1; i >= 0 && v_entry(w, i) >= 0; i--)
    ;
  return i;
}

static void find_dominant(vector *dst, vector *w)
{
  int i;
  v_copy(dst, w);
  while ((i = find_neg_index(dst)) >= 0)
    v_add_multiple(dst, - v_entry(dst, i), afn->cartan_rows[i]);
}

static void calc_mult(weight_set *ws);

static int calc_sum(vector *w, vector *xiw)
{
  int n = v_length(w);
  int result = 0;
  int cont = 1;
  vector *wplus = v_new(n);
  vector *wperm = v_new(n);
  weight_set *ws;
  v_copy(wplus, w);
  
  while (cont)
    {
      v_add_multiple(wplus, 1, xiw);
      find_dominant(wperm, wplus);
      ws = ws_find(hw_set, wperm);
      if (ws != NULL)
	{
	  calc_mult(ws);
	  result += ws_mult(ws) * weight_dot(wplus, xiw, afn);
	}
      else
	cont = 0;
    }
  
  v_free(wplus);
  v_free(wperm);
  
  return result;
}

static int c_value(vector *w)
{
  vector *tmp = v_new(v_length(w));
  int result;
  v_add(tmp, w, afn->rho);
  result = weight_dot(tmp, tmp, afn);
  v_free(tmp);
  return result;
}

struct xi {
  int root_num;
  int o_size;
};

static struct xi *xi_new(int num, int osz)
{
  struct xi *res = weyl_malloc(sizeof(struct xi));
  res->root_num = num;
  res->o_size = osz;
  return res;
}

static void free_xi_list(list *xilist)
{
  int i;
  for (i = 0; i < l_length(xilist); i++)
    weyl_free(l_element(xilist, i));
  l_free(xilist);
}

static list *calc_xi_list(vector *w)
{
  list *xilist = l_new();
  int n = v_length(w);
  int i, r1, r2;
  
  reset_arrays(father_arr, size_arr, dom_arr, nplus);
  
  for (i = 0; i < n; i++)
    if (v_entry(w, i) == 0)
      {
	for (r1 = 0; r1 < nplus; r1++)
	  {
	    vector *root;
	    r2 = action_tab[n * r1 + i];
	    join_roots(r1, r2);
	    root = l_element(afn->weight_roots, r1);
	    if (v_entry(root, i) < 0)
	      dom_arr[r1] = 1;
	  }
      }
  
  for (r1 = 0; r1 < nplus; r1++)
    if (dom_arr[r1] == 0)
      {
	int o_size = 2 * (1 + size_arr[find_father(r1)]);
	l_appendto(xilist, xi_new(r1, o_size));
      }
  
  return xilist;
}

static void calc_mult(weight_set *ws)
{
  vector *w = ws_weight(ws);
  list *xilist;
  int j;
  
  if (ws_mult(ws) != 0)
    return;
  
  xilist = calc_xi_list(w);
  
  /* print_vec(w, 0); */
  ws_mult(ws) = 0;
  for (j = 0; j < l_length(xilist); j++)
    {
      struct xi *xi = l_element(xilist, j);
      int xi_num = xi->root_num;
      int o_size = xi->o_size;
      vector *xiw = l_element(afn->weight_roots, xi_num);
      ws_mult(ws) += o_size * calc_sum(w, xiw);
    }
  
  {
    int c1 = c_value(high_weight);
    int c2 = c_value(w);
    ws_mult(ws) /= (c1 - c2);
  }
  
  free_xi_list(xilist);
}

weight_set *freudenthal(vector *hi_weight, affine *affine, int *actn_tab)
{
  int i, n;
  list *plus_roots;
  list *omega;
  
  afn = affine;
  n = m_dim(afn->cartan);
  plus_roots = afn->positive_roots;
  nplus = l_length(plus_roots);
  action_tab = actn_tab;
  high_weight = hi_weight;
  
  hw_set = calc_weights_below(high_weight, afn);
  omega = ws_to_list(hw_set);
  
  father_arr = (int *) weyl_malloc(nplus * sizeof(int));
  size_arr = (int *) weyl_malloc(nplus * sizeof(int));
  dom_arr = (int *) weyl_malloc(nplus * sizeof(int));
  
  ws_mult(hw_set) = 1;
  for (i = 0; i < l_length(omega); i++)
    calc_mult(l_element(omega, i));
  
  l_free(omega);
  weyl_free(father_arr);
  weyl_free(size_arr);
  weyl_free(dom_arr);
  
  return hw_set;
}
