/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#include "machine.h"
#include "alloc.h"
#include "polynomial.h"
#include "kllib.h"
#include "matrix.h"
#include "tree.h"
#include "calc.h"
#include "hashtab.h"
#include "output.h"


static affine *afn;
static treebase *tb;
static hashtab *kl_cache;

static vector *a;
static vector *xy;

polynomial *kl_poly(node *ny, node *nw);

void new_sum(polynomial *Q, node *ny, node *nz, node *nsw, int s)
{
  int i, lndiff;
  int length, index;
  unsigned mask;
  polynomial *P_y_z, *P_z_sw;
  
  /*  y <= z ?  */
  if (! node_lesseq(ny, nz, tb))
    return;
  
  /*  Have this node been visited before?  */
  length = n_length(nsw);
  index = length / WORD_BIT;
  mask = 1 << (length % WORD_BIT);
  if ((n_colbits(nz)[index] & mask) != 0)
    return;
  n_colbits(nz)[index] |= mask;
  
  /*  Do recursion  */
  for (i = 0; i < n_nbelow(nz); i++)
    new_sum(Q, ny, n_below(nz, i), nsw, s);
  
  /*  Check equal length  */
  lndiff = 1 + n_length(nsw) - n_length(nz);
  if (lndiff % 2 != 0)
    return;
  
  /*  Check s z < z  */
  if (s == 0)
    {
      if (v_dot(afn->high_pair, n_alcove(nz)) < afn->p)
	return;
    }
  else
    {
      if (v_entry(n_alcove(nz), s - 1) > 0)
	return;
    }
  
  /*  Check poly degree  */
  P_z_sw = kl_poly(nz, nsw);
  if (p_coef_used(P_z_sw) == 0 || 2 * p_degree(P_z_sw) != lndiff - 2)
    return;
  
  /*  Add the term  */
  P_y_z = kl_poly(ny, nz);
  p_add_multiple(Q, P_y_z, - p_coef(P_z_sw, (lndiff - 2) / 2), lndiff / 2);
}

void new_unmark(node *nz, int length)
{
  int i;
  int index;
  unsigned mask;
  
  /*  Have this node been visited before?  */
  index = length / WORD_BIT;
  mask = 1 << (length % WORD_BIT);
  if ((n_colbits(nz)[index] & mask) == 0)
    return;
  n_colbits(nz)[index] &= ~mask;
  
  /*  Do recursion  */
  for (i = 0; i < n_nbelow(nz); i++)
    new_unmark(n_below(nz, i), length);
}

void find_min_rep(vector *y, vector *w)
{
  int i, k, n, nperm;
  vector *wt, *perm[6]; /* at most 6 perms of any Dynkin diagram */
  
  n = afn_dim(afn);
  nperm = l_length(afn->permutations);
  
  for (i = 0; i < nperm; i++)
    perm[i] = l_element(afn->permutations, i);
  
  for (k = 0, wt = w  ;  k <= 1  ;  k++, wt = y)
    for (i = 0; i < n; i++)
      {
	int min, x, pi, pn;
	min = v_entry(wt, v_entry(perm[0], i) - 1);
	
	pn = 1;
	for (pi = 1; pi < nperm; pi++)
	  {
	    vector *prm = perm[pi];
	    x = v_entry(wt, v_entry(prm, i) - 1);
	    if (x < min)
	      {
		perm[0] = prm;
		min = x;
		pn = 1;
	      }
	    else if (x == min)
	      {
		perm[pn++] = prm;
	      }
	  }
	nperm = pn;
	if (nperm == 1)
	  break;
      }
  
  for (i = 0; i < n; i++)
    {
      int j = v_entry(perm[0], i) - 1;
      v_entry(xy, i) = v_entry(w, j);
      v_entry(xy, i + n) = v_entry(y, j);
    }
#if 0
  print_vec(xy, 0);
#endif
}

polynomial *kl_poly(node *ny, node *nw)
{
  polynomial *P, *P_sy_sw, *P_y_sw;
  vector *w, *y;
  int s, t, i1, i2;
  node *nsw, *nsy;
  void **polyp;
  int n = tb_dim(tb);
  
  /* check for trivial cases */
  if (ny == nw)
    return tb_one(tb);
  if (! node_lesseq(ny, nw, tb))
    return tb_zero(tb);
  if (n_length(nw) - n_length(ny) <= 2)
    return tb_one(tb);
  
  w = n_alcove(nw);
  y = n_alcove(ny);
  
  /* create key (xy) for kl_cache */
  if (l_length(afn->permutations) == 1)
    {
      int i;
      for (i = 0; i < n; i++)
	{
	  v_entry(xy, i) = v_entry(w, i);
	  v_entry(xy, i + n) = v_entry(y, i);
	}
    }
  else
    find_min_rep(y, w);
  
  /* lookup key */
  polyp = hash_mkfind(kl_cache, xy);
  if (*polyp != NULL)
    return *polyp;
  
  P = p_new(0);
  *polyp = P;
  
  /* xy has been absorbed in hash table, make new key. */
  xy = v_new(2 * n);
  
  s = 0;
  /* choose simple reflexion s, so  s w < w.  Try for s y > y */
  for (t = n; t > 0; t--)
    if (v_entry(w, t - 1) < 0)
      {
	s = t;
	if (v_entry(y, t - 1) >= 0)
	  break;
      }
#if 0
  /* try harder for  s y > y.  Seems to be slower. */
  if ((s != t) && (v_dot(w, afn->high_pair) >= afn->p))
    s = 0;
#endif

  /* calculate  s w  */
  v_copy(a, w);
  ref_weight(a, s, afn);
  nsw = tb_findalc(tb, a);
  
  /* calculate  s y  */
  v_copy(a, y);
  ref_weight(a, s, afn);
  nsy = tb_findalc(tb, a);
  
  /* calculate polynomial */
  new_sum(P, ny, nsw, nsw, s);
  new_unmark(nsw, n_length(nsw));
  
  P_sy_sw = kl_poly(nsy, nsw);
  P_y_sw = kl_poly(ny, nsw);
  
  i1 = i2 = 0;
  if (n_length(nsy) < n_length(ny))
    i1 = 1;
  else
    i2 = 1;
  
  p_add_multiple(P, P_y_sw, 1, i1);
  p_add_multiple(P, P_sy_sw, 1, i2);
  
  return P;
}

polynomial *kl_polynomial(list *y, list *w, treebase *_tb)
{
  node *ny, *nw;
  polynomial *P;
  
  tb = _tb;
  if (tb_kl_cache(tb) == NULL)
    {
      tb_kl_cache(tb) = hash_new((cmp_t) v_compare, (hash_t) v_hash);
      tb_zero(tb) = p_new(0);
      tb_one(tb) = p_new(1);
    }
  kl_cache = tb_kl_cache(tb);
  afn = tb_affine(tb);
  a = v_new(tb_dim(tb));
  xy = v_new(2 * tb_dim(tb));
  
  calc_alcove(a, y, afn);
  ny = tb_findalc(tb, a);
  
  calc_alcove(a, w, afn);
  nw = tb_findalc(tb, a);
  
  if (ny != NULL && nw != NULL)
    P = kl_poly(ny, nw);
  else
    P = tb_zero(tb);
  
  v_free(a);
  v_free(xy);
  
  return P;
}

void free_kl_cache(treebase *tb)
{
  hashtab *ht = tb_kl_cache(tb);
  hash_itr itr;
  
  for (hash_first(ht, itr); hash_good(itr); hash_next(itr))
    {
      v_free(hash_key(itr));
      p_free(hash_value(itr));
    }
  hash_free(ht);
  
  p_free(tb_zero(tb));
  p_free(tb_one(tb));
}
