/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "alloc.h"
#include "list.h"
#include "matrix.h"
#include "affine.h"
#include "output.h"
#include "calc.h"
#include "tree.h"
#include "kllib.h"
#include "weight.h"
#include "freudlib.h"
#include "cmdline.h"


void print_usage(char *s)
{
  fprintf(stderr, "Usage: %s [-p] rsys-p weight\n", s);
  exit(1);
}

int main(int ac, char **av)
{
  treebase *tb;
  list *w0, *w, *w0w, *w0y;
  node *ny;
  polynomial *P;
  vector *a, *high_weight, *lambda;
  int n, i, j, l, warn;
  int *action_tab;
  weight_set *result, *ws;
  int opt_print_polys, c;
  affine *afn;
  
  cmd_init(ac, av);
  
  c = getopt(ac, av, "p");
  if (c == '?')
    print_usage(av[0]);
  opt_print_polys = (c == 'p');
  
  afn = cmd_rsys();
  high_weight = cmd_weight(1, afn);
  
  w0 = afn->longest_element;
  n = afn_dim(afn);
  
  warn = 0;
  for (i = 0; i < n; i++)
    if (v_entry(high_weight, i) > afn->p)
      warn = 1;
  if (warn)
    fprintf(stderr, "%s: Warning: weight is not restricted!\n", av[0]);
  
  lambda = v_new(n);
  w = l_new();
  if (weight_to_exp(w, high_weight, lambda, afn))
    fprintf(stderr, "%s: Warning: weight belongs to edge!\n", av[0]);
  
  w0w = l_new();
  l_copy(w0w, w0);
  for (i = 0; i < l_length(w); i++)
    l_appendto(w0w, l_element(w, i));
  reduced_exp(w0w, w0w, afn);
  
  /* FIXME: not sure about 2nd arg */
  tb = build_tree(w0w, 0, afn);
  
  calc_bit_matrices(tb);
  tb_init_colbits(tb);
  
  action_tab = build_action_tab(afn);
  
  a = v_new(tb_dim(tb));
  w0y = l_new();
  
  result = ws_new(lambda);
  
  for (l = 0; l <= tb_max_length(tb); l++)
    for (ny = tb_node_list(tb, l); ny != NULL; ny = n_next(ny))
      {
	v_copy(a, n_alcove(ny));
	apply_to_alcove(a, w0, afn);
	
	for (i = v_length(a); i > 0 && v_entry(a, i - 1) > 0; i--)
	  ;
	if (i == 0)
	  {
	    int p1;
	    alcove_to_exp(w0y, a, afn);
	    P = kl_polynomial(w0y, w0w, tb);
	    
	    if (opt_print_polys)
	      {
		int k;
		printf("P(");
		for (k = 0; k < v_length(a); k++)
		  printf("%d", v_entry(a, k) - 1);
		printf(", ");
		for (k = 0; k < v_length(a); k++)
		  printf("%d", v_entry(high_weight, k) - 1);
		printf(") = ");
		print_poly(P);
	      }
	    
	    p1 = 0;
	    for (j = 0; j <= p_degree(P); j++)
	      p1 += p_coef(P, j);
	    if ((l_length(w) - l_length(w0y)) % 2 == 1)
	      p1 = - p1;
	    
	    v_copy(a, lambda);
	    apply_to_alcove(a, w0y, afn);
	    v_sub(a, a, afn->rho);
	    ws = freudenthal(a, afn, action_tab);
	    ws_add_multiple(result, p1, ws);
	    ws_free(ws);
	  }
      }
  
  if (opt_print_polys)
    putchar('\n');
  
  print_weight_set(result, 1, 0);
  
  ws_free(result);
  v_free(high_weight);
  v_free(lambda);
  v_free(a);
  l_free(w);
  l_free(w0w);
  l_free(w0y);
  tb_free(tb);
  weyl_free(action_tab);
  
  memory_report;
  
  exit(0);
}
