/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#include "alloc.h"
#include "matrix.h"
#include "output.h"
#include "list.h"
#include "calc.h"
#include "tree.h"
#include "polynomial.h"
#include "weight.h"

void fprint_intlist(FILE *s, list *l, int indent)
{
  int i;
  for (i = 0; i < indent; i++)
    putc(' ', s);
  putc('{', s);
  for (i = 0; i < l_length(l); i++)
    fprintf(s, (i == 0) ? "%d" : ", %d", (int) l_element(l, i));
  fprintf(s, "}\n");
}

void fprint_vec(FILE *s, vector *v, int indent)
{
  int i;
  for (i = 0; i < indent; i++)
    putc(' ', s);
  putc('(', s);
  for (i = 0; i < v_length(v); i++)
    fprintf(s, (i == 0) ? "%d" : ", %d", (int) v_entry(v, i));
  fprintf(s, ")\n");
}

void fprint_mat(FILE *s, matrix *m)
{
  int n = m_dim(m);
  int i, j;
  
  for (i = 0; i < n; i++)
    {
      for (j = 0; j < n; j++)
	fprintf(s, (j == 0) ? "%4d" : " %4d", m_entry(m, i, j));
      putc('\n', s);
    }
}

void fprint_affine(FILE *s, affine *afn)
{
  int i;
  fprintf(s, "Root system: %c%d\n", afn->root_system_type, m_dim(afn->cartan));
  
  fprintf(s, "\ncartan matrix:\n");
  fprint_mat(s, afn->cartan);
  
  fprintf(s, "\ntransposed of cartan matrix:\n");
  fprint_mat(s, afn->tcartan);
  
  fprintf(s, "\ndeterminant of cartan matrix: %d \n", afn->dcartan);
  
  fprintf(s, "\nadjoint of cartan matrix:\n");
  fprint_mat(s, afn->acartan);
  
  fprintf(s, "\nadjoint of transposed cartan matrix:\n");
  fprint_mat(s, afn->atcartan);
  
  fprintf(s, "\nsquares of lengths of simple roots:\n");
  for (i = 0; i < m_dim(afn->cartan); i++)
    fprintf(s, (i == 0) ? "   %d" : ", %d", afn->simple_length[i]);
  
  fprintf(s, "\n\n2 * inner products of simple roots:\n");
  fprint_mat(s, afn->root_dot);
  
  fprintf(s, "\n%d * inner products of fundamental weights:\n",
	  afn->wdot_factor);
  fprint_mat(s, afn->weight_dot);
  
  fprintf(s, "\ncoxeter number: h = %d\n", afn->h);
  fprintf(s, "prime number:   p = %d\n", afn->p);
  
  fprintf(s, "\nweyl group element for top alcove:\n");
  fprint_intlist(s, afn->top_alcove, 3);
  
  fprintf(s, "\nlongest element of weyl group:\n");
  fprint_intlist(s, afn->longest_element, 3);
  
  fprintf(s, "\npositive roots:\n");
  {
    list *plusroots = afn->positive_roots;
    for (i = 0; i < l_length(plusroots); i++)
      fprint_vec(s, l_element(plusroots, i), 3);
  }
  
  fprintf(s, "\nlengths of positive roots:\n");
  fprint_intlist(s, afn->root_length, 3);
  
  fprintf(s, "\npermutations of dynkin diagram:\n");
  for (i = 0; i < l_length(afn->permutations); i++)
    fprint_vec(s, l_element(afn->permutations, i), 3);
  
  fprintf(s, "\nhighest root:\n");
  fprint_vec(s, afn->highest_root, 3);
  
  fprintf(s, "\nhighest_simp = %d\n", afn->highest_simp);
  fprintf(s, "highest_elem =\n");
  fprint_mat(s, afn->highest_elem);
  
  fprintf(s, "\nhigh_pair =\n");
  fprint_vec(s, afn->high_pair, 3);
  
  fprintf(s, "\nhighest root in weight coordinates:\n");
  fprint_vec(s, afn->high_weight, 3);
  
  fprintf(s, "\nrho:\n");
  fprint_vec(s, afn->rho, 3);
  
  fprintf(s, "\nrows in cartan matrix:\n");
  for (i = 0; i < m_dim(afn->cartan); i++)
    fprint_vec(s, afn->cartan_rows[i], 3);

  fprintf(s, "\ncolumns in cartan matrix:\n");
  for (i = 0; i < m_dim(afn->cartan); i++)
    fprint_vec(s, afn->cartan_columns[i], 3);
}

void fprint_table(FILE *s, list *vec_list, matrix *m)
{
  int nvec = l_length(vec_list);
  vector *first = l_element(vec_list, 0);
  int n = v_length(first);
  int i, j;
  
  for (i = 0; i < n; i++)
    {
      for (j = 0; j <= n + 4; j++)
	fprintf(s, " ");
      
      for (j = 0; j < nvec; j++)
	{
	  vector *root = l_element(vec_list, j);
	  fprintf(s, "%3d", v_entry(root, i));
	}
      fprintf(s, "\n");
    }
  fprintf(s, "\n");
  
  for (i = 0; i < nvec; i++)
    {
      vector *a = l_element(vec_list, i);
      fprintf(s, "%2d  ", i + 1);
      for (j = 0; j < n; j++)
	fprintf(s, "%d", v_entry(a, j));
      fprintf(s, " ");
      
      for (j = 0; j < nvec; j++)
	fprintf(s, "%3d", m_entry(m, i, j));
      fprintf(s, "\n");
    }
}


void fprint_node(FILE *s, node *nd, int mode, affine *afn)
{
  list *exp = NULL;
  vector *a = NULL;
  int i;
  
  if (mode == PTM_WEYL)
    exp = l_new();
  else if (mode == PTM_WEIGHT)
    a = v_new(afn_dim(afn));

  if (mode == PTM_WEYL)
    {
      alcove_to_exp(exp, n_alcove(nd), afn);
      fprint_intlist(s, exp, 0);
    }
  else if (mode == PTM_WEIGHT)
    {
      v_sub(a, n_alcove(nd), afn->rho);
      fprint_vec(s, a, 0);
    }
  else if (mode == PTM_COORD)
    {
      fprint_vec(s, n_alcove(nd), 0);
    }
  
  for (i = 0; i < n_nbelow(nd); i++)
    {
      vector *alc = n_alcove(n_below(nd, i));
      
      if (mode == PTM_WEYL)
	{
	  alcove_to_exp(exp, alc, afn);
	  fprint_intlist(s, exp, 2);
	}
      else if (mode == PTM_WEIGHT)
	{
	  v_sub(a, alc, afn->rho);
	  fprint_vec(s, a, 2);
	}
      else if (mode == PTM_COORD)
	{
	  fprint_vec(s, alc, 2);
	}
    }
  putc('\n', s);
  
  if (mode == PTM_WEYL)
    l_free(exp);
  else if (mode == PTM_WEIGHT)
    v_free(a);
}

static void _pnb(FILE *s, node *nd, int mode, affine *afn)
{
  int i;
  
  if (n_color(nd) != 0)
    return;
  n_color(nd) = 1;
  
  fprint_node(s, nd, mode, afn);
  
  for (i = 0; i < n_nbelow(nd); i++)
    _pnb(s, n_below(nd, i), mode, afn);
}

void fprint_nodes_below(FILE *s, treebase *tb, node *nd, int mode)
{
  tb_reset_node_colors(tb);
  _pnb(s, nd, mode, tb_affine(tb));
}

void fprint_tree(FILE *s, treebase *tb, int mode)
{
  int l;
  node *nd;
  for (l = tb_max_length(tb); l >= 0; l--)
    for (nd = tb_node_list(tb, l); nd != NULL; nd = n_next(nd))
      fprint_node(s, nd, mode, tb_affine(tb));
}

void fprint_term(FILE *s, int coef, int deg, int first)
{
  if (! first)
    fputs(coef < 0 ? " - " : " + ", s);
  else if (coef < 0)
    putc('-', s);
  
  coef = abs(coef);
  
  if (deg == 0)
    {
      fprintf(s, "%d", coef);
      return;
    }
  
  if (coef > 1)
    fprintf(s, "%d ", coef);
  putc('q', s);
  if (deg > 1)
    fprintf(s, "^%d", deg);
}

void fprint_poly(FILE *s, polynomial *p)
{
  if (p_coef_used(p) == 0)
    fprintf(s, "0\n");
  else
    {
      int i;
      int first = 1;
      for (i = 0; i < p_coef_used(p); i++)
	if (p->coef_list[i] != 0)
	  {
	    fprint_term(s, p->coef_list[i], i, first);
	    first = 0;
	  }
      putc('\n', s);
    }
}

void fprint_weight(FILE *s, vector *w, int add)
{
  int j;
  for (j = 0; j < v_length(w); j++)
    printf("%2d", v_entry(w, j) + add);
}

void fprint_weight_set(FILE *s, weight_set *ws, int prtcoef, int add)
{
  if (ws == NULL)
    return;
  fprint_weight_set(s, ws_left(ws), prtcoef, add);
  if (prtcoef != 1 || ws_mult(ws) != 0)
    {
      vector *w = ws_weight(ws);
      print_weight(w, add);
      if (prtcoef)
	fprintf(s, " :  %d\n", ws_mult(ws));
      else
	fprintf(s, "\n");
    }
  fprint_weight_set(s, ws_right(ws), prtcoef, add);
}

void fprint_vec_list(FILE *s, list *veclist)
{
  int i;
  for (i = 0; i < l_length(veclist); i++)
    fprint_vec(s, l_element(veclist, i), 0);
}

void fprint_lspath(FILE *s, list *pi)
{
  int ln = l_length(pi) / 3;
  int i;
  for (i = 0; i < ln; i++)
    {
      fprintf(s, "%d/%d :", 
	     l_intelem(pi, 3 * i + 1), 
	     l_intelem(pi, 3 * i + 2));
      fprint_vec(s, l_element(pi, 3 * i), 2);
    }
}
