/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "efhtree.h"
#include "redulib2.h"
#include "alloc.h"
#include "sconst.h"
#include "weight.h"
#include "list.h"
#include "rank.h"

static list *build_lambda_pairs(vector *lambda, affine *afn)
{
  list *plus_roots = afn->positive_roots;
  list *root_length = afn->root_length;
  int nroots = l_length(plus_roots);
  int n = m_dim(afn->cartan);
  list *res = l_new();
  int j, p = afn->p;
  
  vector *lr = v_new(n);
  vector *tmp = v_new(n);
  m_multvec(tmp, afn->atcartan, lambda);
  m_vecmult(lr, tmp, afn->root_dot);
  
  for (j = 0; j < nroots; j++)
    {
      vector *aj = l_element(plus_roots, j);
      l_appendint
	(res, (v_dot(lr, aj) / (afn->dcartan * l_intelem(root_length, j)))% p);
    }
  
  v_free(tmp);
  v_free(lr);
  return res;
}

static matrix *Nneg;
static matrix *sub_mat;
static matrix *pair_mat;
static list *lambda_pairs;
int prime;


void init_redulib2(vector *lambda, affine *afn)
{
  matrix *Npos, *add_mat;
  list *plus_roots = afn->positive_roots;
  prime = afn->p;
  
  Nneg = m_new(l_length(plus_roots));
  sub_mat = build_add_matrix(plus_roots, -1);
  Npos = m_new(l_length(plus_roots));
  add_mat = build_add_matrix(plus_roots, 1);
  calc_sconsts(afn, add_mat, sub_mat, Npos, Nneg);
  m_free(Npos);
  m_free(add_mat);
  
  pair_mat = build_pair_matrix(afn);
  lambda_pairs = build_lambda_pairs(lambda, afn);
  
  reduce_mat_mod_p(Nneg, prime);
  reduce_mat_mod_p(pair_mat, prime);
}

void free_redulib2()
{
  m_free(Nneg);
  m_free(sub_mat);
  m_free(pair_mat);
  l_free(lambda_pairs);
}


list *efh_trees;

static int do_reduce(list *efh)
{
  efh_tree *node, *t;
  int root;
  int i, x, ln = l_length(efh);
  
  if (ln == 0)
    {
      l_free(efh);
      return 1;
    }
  
  if (efh_is_f(l_intelem(efh, 0))
      || efh_is_e(l_intelem(efh, ln - 1)))
    {
      l_free(efh);
      return 0;
    }
  
  t = l_element(efh_trees, ln);
  if (t == NULL)
    node = l_element(efh_trees, ln) = efh_new(efh);
  else if (efh_insert(t, efh, &node))
    {
      l_free(efh);
      return efh_value(node);
    }
  
  i = ln;
  while (i > 0 && efh_is_f(l_intelem(efh, --i)))
    ;
  
  x = l_intelem(efh, i);
  root = efh_i(x);
  
  if (efh_is_h(x))
    {
      if (i == ln - 1)
	{
	  int c = l_intelem(lambda_pairs, root - 1);
	  if (c != 0)
	    {
	      l_length(efh)--;
	      c *= do_reduce(efh);
	    }
	  else
	    l_free(efh);
	  return efh_value(node) = c % prime;
	}
      else
	{
	  int nxt_root = efh_i(l_intelem(efh, i + 1));
	  int pair = m_entry(pair_mat, nxt_root - 1, root - 1);
	  int tmp = 0;
	  
	  if (pair != 0)
	    {
	      list *efh2 = l_new_copy(efh);
	      l_remove_elem(efh2, i);
	      tmp = pair * do_reduce(efh2);
	    }
	  
	  l_intelem(efh, i) = efh_mk_f(nxt_root);
	  l_intelem(efh, i + 1) = efh_mk_h(root);
	  return node->value =
	    (do_reduce(efh) - tmp) % prime;
	}
    }
  else  /* efh_is_e(x) */
    {
      int nxt_root = efh_i(l_intelem(efh, i + 1));
      if (root == nxt_root)
	{
	  list *efh2 = l_new_copy(efh);
	  l_intelem(efh2, i) = efh_mk_f(nxt_root);
	  l_intelem(efh2, i + 1) = efh_mk_e(root);
	  
	  l_remove_elem(efh, i);
	  l_intelem(efh, i) = efh_mk_h(root);
	  
	  return node->value =
	    (do_reduce(efh) + do_reduce(efh2)) % prime;
	}
      else /* root != nxt_root */
	{
	  int tmp = m_entry(Nneg, root - 1, nxt_root - 1);
	  
	  if (tmp != 0)
	    {
	      int dif = m_entry(sub_mat, root - 1, nxt_root - 1);
	      list *efh2 = l_new_copy(efh);
	      l_remove_elem(efh2, i);
	      l_intelem(efh2, i) = (dif > 0) ? efh_mk_e(dif) : efh_mk_f(-dif);
	      tmp *= do_reduce(efh2);
	    }
	  
	  l_intelem(efh, i) = efh_mk_f(nxt_root);
	  l_intelem(efh, i + 1) = efh_mk_e(root);
	  
	  return node->value = 
	    (do_reduce(efh) + tmp) % prime;
	}
    }
}

static list *sums_to_efhlist(vector *sum1, vector *sum2)
{
  list *efh = l_new();
  int i, j;
  for (i = v_length(sum1) - 1; i >= 0; i--)
    {
      for (j = 0; j < v_entry(sum1, i); j++)
	l_appendint(efh, efh_mk_e(i + 1));
    }
  for (i = 0; i < v_length(sum2); i++)
    {
      for (j = 0; j < v_entry(sum2, i); j++)
	l_appendint(efh, efh_mk_f(i + 1));
    }
  return efh;
}

int reduce2(vector *sum1, vector *sum2)
{
  list *efh = sums_to_efhlist(sum1, sum2);
  int i, res;
  
  efh_trees = l_new_copy(efh);
  l_appendto(efh_trees, NULL);
  for (i = 0; i < l_length(efh_trees); i++)
    l_element(efh_trees, i) = NULL;
  
  res = do_reduce(efh);
  
  for (i = 0; i < l_length(efh_trees); i++)
    efh_free(l_element(efh_trees, i));
  l_free(efh_trees);
  
  return res < 0 ? res + prime : res;
}
