/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "alloc.h"
#include "list.h"
#include "matrix.h"
#include "affine.h"
#include "sconst.h"

#ifdef USE_EKS_NMAT
#include <ctype.h>
#include "output.h"
#endif

int find_root(list *plus_roots, vector *a)
{
  int i, sign;
  
  i = 0;
  while (i < v_length(a) && v_entry(a, i) == 0)
    i++;
  if (i == v_length(a))
    return 0;
  else if (v_entry(a, i) > 0)
    sign = 1;
  else
    {
      v_scale(a, -1, a);
      sign = -1;
    }
  
  i = l_length(plus_roots);
  while (i > 0 && v_compare(a, l_element(plus_roots, i - 1)) != 0)
    i--;
  return sign * i;
}

matrix *build_add_matrix(list *plus_roots, int addsub)
{
  int i, j, n, r;
  int nroots = l_length(plus_roots);
  matrix *res = m_new(nroots);
  vector *a;
  
  a = l_element(plus_roots, 0);
  n = v_length(a);
  a = v_new(n);
  
  for (i = 0; i < nroots; i++)
    {
      m_entry(res, i, i) = 0;
      
      for (j = 0; j < i; j++)
	{
	  if (addsub == -1)
	    v_sub(a, l_element(plus_roots, i), l_element(plus_roots, j));
	  else
	    v_add(a, l_element(plus_roots, i), l_element(plus_roots, j));
	  r = find_root(plus_roots, a);
	  m_entry(res, i, j) = r;
	  m_entry(res, j, i) = addsub * r;
	}
    }
  v_free(a);
  return res;
}  

matrix *build_pair_matrix(affine *afn)
{
  list *plus_roots = afn->positive_roots;
  list *root_length = afn->root_length;
  int nroots = l_length(plus_roots);
  int n = m_dim(afn->cartan);
  vector *tmpvec = v_new(n);
  matrix *rdot = afn->root_dot;
  matrix *res = m_new(nroots);
  int i, j;
  
  for (i = 0; i < nroots; i++)
    for (j = 0; j < nroots; j++)
      {
	vector *ai = l_element(plus_roots, i);
	vector *aj = l_element(plus_roots, j);
	m_multvec(tmpvec, rdot, aj);
	m_entry(res, i, j) = v_dot(ai, tmpvec) / l_intelem(root_length, j);
      }
  
  v_free(tmpvec);
  return res;
}

static matrix *N;
static matrix *add_mat, *sub_mat;
static int *sum_list1;
static int *sum_list2;
static list *root_length;

static int calc_n_abs(int i, int j)
{
  int t = 0;
  
  while (j != 0)
    {
      t++;
      if (j > 0)
	j = m_entry(sub_mat, j - 1, i - 1);
      else
	j = - m_entry(add_mat, -j - 1, i - 1);
    }
  
  return t;
}

static int calc_n_entry(int i, int j);

static int calc_n_negind(int i, int j)
{
  int k;
  
  k = m_entry(sub_mat, -j - 1, i - 1);
  
  if (k == 0)
    {
      return 0;
    }
  else if (k > 0)
    {
      return calc_n_entry(k, i) * l_intelem(root_length, k - 1)
	/ l_intelem(root_length, -j - 1);
    }
  else /* k < 0 */
    {
      return calc_n_entry(-k, -j) * l_intelem(root_length, -k - 1)
	/ l_intelem(root_length, i - 1);
    }
}

static int calc_n_entry(int i, int j)
{
  int sum_root, res;
  
  sum_root = m_entry(add_mat, i - 1, j - 1);
  if (sum_root == 0)
    return 0;
  
  if (m_entry(N, i - 1, j - 1) != 0)
    return m_entry(N, i - 1, j - 1);
  
  if (sum_list1[sum_root - 1] == 0)
    {
      /* we can choose the sign freely */
      sum_list1[sum_root - 1] = i;
      sum_list2[sum_root - 1] = j;
      res = calc_n_abs(i, j);
    }
  else
    {
      /* calculate N(i,j) from other entries */
      int k = sum_list1[sum_root - 1];
      int l = sum_list2[sum_root - 1];
      int r, sqr_ij, sqr_ik, sqr_jk;
      int n_kl, n_jk, n_il, n_ki, n_jl;
      
      sqr_ij = l_intelem(root_length, sum_root - 1);
      r = m_entry(sub_mat, i - 1, k - 1);
      sqr_ik = (r == 0) ? 1 : l_intelem(root_length, abs(r) - 1);
      r = m_entry(sub_mat, j - 1, k - 1);
      sqr_jk = (r == 0) ? 1 : l_intelem(root_length, abs(r) - 1);
      
      n_kl = m_entry(N, k - 1, l - 1);
      n_jk = calc_n_negind(j, -k);
      n_il = calc_n_negind(i, -l);
      n_ki = - calc_n_negind(i, -k);
      n_jl = calc_n_negind(j, -l);
      
      res = sqr_ij * (sqr_ik * n_jk * n_il + sqr_jk * n_ki * n_jl)
	/ (n_kl * sqr_jk * sqr_ik);
    }
  
  m_entry(N, i - 1, j - 1) = res;
  m_entry(N, j - 1, i - 1) = - res;
  
  return res;
}

void calc_sconsts(affine *afn, matrix *addmat, matrix *submat,
		     matrix *Npos, matrix *Nneg)
{
  int i, j;
  int nplus;
  
  root_length = afn->root_length;
  nplus = l_length(root_length);
  
  add_mat = addmat;
  sub_mat = submat;
  
  N = Npos;
  m_set_zero(N);
  
  sum_list1 = weyl_calloc(2 * nplus, sizeof(int));
  sum_list2 = sum_list1 + nplus;
  
  for (i = 1; i <= nplus; i++)
    for (j = 1; j <= i; j++)
      {
	int c;
	calc_n_entry(i, j);
	c = calc_n_negind(i, -j);
	m_entry(Nneg, i - 1, j - 1) = c;
	m_entry(Nneg, j - 1, i - 1) = c;
      }
  
  weyl_free(sum_list1);
  
#ifdef USE_EKS_NMAT
  printf("Result:\n");
  print_table(afn->positive_roots, N);
  if (eks_mat != NULL)
    m_free(eks_mat);
#endif
}

#ifdef __GNUC__
inline
#endif
int get_n_value(matrix *Npos, matrix *Nneg, int i, int j)
{
  if (i > 0)
    if (j > 0)
      return m_entry(Npos, i - 1, j - 1);
    else
      return m_entry(Nneg, i - 1, -j - 1);
  else
    if (j > 0)
      return m_entry(Nneg, -i - 1, j - 1);
    else
      return m_entry(Nneg, -j - 1, -i - 1);
}

#ifdef __GNUC__
inline
#endif
int add_roots(matrix *addmat, matrix *submat, int i, int j)
{
  if (i > 0)
    if (j > 0)
      return m_entry(addmat, i - 1, j - 1);
    else
      return m_entry(submat, i - 1, -j - 1);
  else
    if (j > 0)
      return m_entry(submat, j - 1, -i - 1);
    else
      return - m_entry(addmat, -i - 1, -j - 1);
}

#ifdef __GNUC__
inline
#endif
int get_pair(matrix *pair_mat, int i, int j)
{
  int pair = m_entry(pair_mat, abs(i) - 1, abs(j) - 1);
  if (i < 0)
    pair = - pair;
  if (j < 0)
    pair = - pair;
  return pair;
}
