/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "machine.h"
#include "alloc.h"
#include "matrix.h"
#include "tree.h"
#include "polynomial.h"
#include "calc.h"
#include "kllib.h"

#define DEFAULT_NODE_COUNT_SZ	40

node *n_new(int length, vector *alc)
{
  node *n = (node *) weyl_malloc(sizeof(node) + (length - 1) * sizeof(node *));
  n->length = length;
  n->alcove = alc;
  n->nbelow = 0;
  return n;
}

void n_free(node *nd)
{
  v_free(n_alcove(nd));
  weyl_free(nd);
}


void tb_reset_node_colors(treebase *tb)
{
  int l;
  node *nd;
  for (l = 0; l <= tb_max_length(tb); l++)
    for (nd = tb_node_list(tb, l); nd != NULL; nd = n_next(nd))
      n_color(nd) = 0;
}


void tb_init_colbits(treebase *tb)
{
  int i, num_nodes, l;
  unsigned *bits;
  node *nd;
  int max_length = tb_max_length(tb);
  int vec_length = (max_length + WORD_BIT - 1) / WORD_BIT;
  
  num_nodes = 0;
  for (i = 0; i <= tb_max_length(tb); i++)
    num_nodes += tb_node_count(tb, i);
  
  bits = (unsigned *) 
    weyl_calloc(num_nodes * vec_length, sizeof(unsigned));
  tb_all_colbits(tb) = bits;
  
  for (l = 0; l <= tb_max_length(tb); l++)
    for (nd = tb_node_list(tb, l); nd != NULL; nd = n_next(nd))
      {
	n_colbits(nd) = bits;
	bits += vec_length;
      }
}

treebase *tb_new(affine *afn)
{
  treebase *tb = (treebase *) 
    weyl_calloc(1, sizeof(treebase));
  
  tb->dim = m_dim(afn->cartan);
  tb->afn = afn;
  tb->nodeset = hash_new((cmp_t) v_compare, (hash_t) v_hash);
  tb->max_length = 0;
  tb->node_count_sz = DEFAULT_NODE_COUNT_SZ;
  tb->node_count = (int *) weyl_calloc(1, tb->node_count_sz * sizeof(int));
  tb->node_list = (node **) weyl_calloc(1, tb->node_count_sz * sizeof(node *));
  
  return tb;
}

void tb_set_max_length(treebase *tb, int length)
{
  int sz;
  
  if (length <= tb->max_length)
    return;
  tb->max_length = length;
  
  sz = tb_node_count_sz(tb);
  if (sz <= length)
    {
      int new_sz = 2 * length;
      tb->node_count_sz = new_sz;
      
      tb->node_count = weyl_realloc(tb->node_count, new_sz * sizeof(int));
      memset(tb->node_count + sz, 0, (new_sz - sz) * sizeof(int));
      
      tb->node_list = weyl_realloc(tb->node_list, new_sz * sizeof(node *));
      memset(tb->node_list + sz, 0, (new_sz - sz) * sizeof(node *));
    }
}

void tb_free(treebase *tb)
{
  node *nd, *next;
  int l;
  for (l = 0; l <= tb_max_length(tb); l++)
    for (nd = tb_node_list(tb, l); nd != NULL; nd = next)
      {
	next = n_next(nd);
	n_free(nd);
      }
  
  hash_free(tb_nodeset(tb));
  if (tb_affine(tb) != NULL)
    afn_free(tb_affine(tb));
  weyl_free(tb->node_count);
  weyl_free(tb->node_list);
  
  if (tb->bit_matrices != NULL)
    {
      int max_length = tb_max_length(tb);
      int max2 = (max_length - 1) * (max_length - 1);
      int i;
      
      for (i = 0; i < max2; i++)
	if (tb->bit_matrices[i] != NULL)
	  weyl_free(tb->bit_matrices[i]);
      
      weyl_free(tb->bit_matrices);
    }
  
  if (tb->all_colbits != NULL)
    weyl_free(tb->all_colbits);
  
  if (tb->kl_cache != NULL)
    free_kl_cache(tb);
  
  weyl_free(tb);
}

node *tb_find(treebase *tb, list *exp)
{
  int n = tb_dim(tb);
  vector *a = v_new(n);
  node *node;
  calc_alcove(a, exp, tb_affine(tb));
  node = hash_lookup(tb_nodeset(tb), a);
  v_free(a);
  return node;
}
