/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "alloc.h"
#include "calc.h"
#include "matrix.h"
#include "affine.h"
#include "weight.h"

weight_set *ws_new(vector *w)
{
  weight_set *ws = (weight_set *) weyl_calloc(1, sizeof(weight_set));
  ws_weight(ws) = v_new_copy(w);
  return ws;
}

void ws_free_struc(weight_set *ws, int free_weights)
{
  if (ws == NULL)
    return;
  if (free_weights)
    v_free(ws_weight(ws));
  ws_free_struc(ws_left(ws), free_weights);
  ws_free_struc(ws_right(ws), free_weights);
  weyl_free(ws);
}

weight_set *ws_find(weight_set *ws, vector *w)
{
  while (ws != NULL)
    {  
      int c = v_compare(w, ws_weight(ws));
      if (c == 0)
	return ws;
      if (c < 0)
	ws = ws_left(ws);
      else
	ws = ws_right(ws);
    }
  return ws;
}

weight_set *ws_insert(weight_set *ws, vector *w)
{
  weight_set *last = NULL;
  int c = 1;
  while (c != 0 && ws != NULL)
    {
      last = ws;
      c = v_compare(w, ws_weight(ws));
      if (c < 0)
	ws = ws_left(ws);
      else if (c > 0)
	ws = ws_right(ws);
    }
  
  if (c == 0)
    return NULL;
  
  if (c < 0)
    return ws_left(last) = ws_new(w);
  else
    return ws_right(last) = ws_new(w);
}


void ws_add_weight_mult(weight_set *ws, vector *w, int m)
{
  weight_set *last = NULL;
  int c = 1;
  while (c != 0 && ws != NULL)
    {
      last = ws;
      c = v_compare(w, ws_weight(ws));
      if (c < 0)
	ws = ws_left(ws);
      else if (c > 0)
	ws = ws_right(ws);
    }
  
  if (c < 0)
    ws = ws_left(last) = ws_new(w);
  else if (c > 0)
    ws = ws_right(last) = ws_new(w);
  
  ws_mult(ws) += m;
}

void ws_add_multiple(weight_set *dst, int m, weight_set *src)
{
  if (src == NULL)
    return;
  ws_add_multiple(dst, m, ws_left(src));
  ws_add_multiple(dst, m, ws_right(src));
  ws_add_weight_mult(dst, ws_weight(src), m * ws_mult(src));
}


int weight_dot(vector *w1, vector *w2, affine *afn)
{
  int result;
  vector *tmp = v_new(v_length(w1));
  m_multvec(tmp, afn->weight_dot, w2);
  result = v_dot(w1, tmp);
  v_free(tmp);
  return result;
}

weight_set *calc_weights_below(vector *high_weight, affine *afn)
{
  int n = m_dim(afn->cartan);
  vector *lambda = v_new(n);
  weight_set *ws, *hw_set = ws_new(high_weight);
  int j, l;
  list *omega = l_new();
  l_appendto(omega, high_weight);
  
  while (l_length(omega) > 0)
    {
      vector *w = l_remove_last(omega);
      for (j = 0; j < l_length(afn->weight_roots); j++)
	{
	  v_copy(lambda, w);
	  v_sub(lambda, lambda, l_element(afn->weight_roots,  j));
	  for (l = n; l > 0 && v_entry(lambda, l - 1) >= 0; l--)
	    ;
	  if (l == 0 && (ws = ws_insert(hw_set, lambda)) != NULL)
	    l_appendto(omega, ws_weight(ws));
	}
    }
  v_free(lambda);
  l_free(omega);
  return hw_set;
}


void ws_append_to_list(list *wslist, weight_set *ws)
{
  if (ws == NULL)
    return;
  ws_append_to_list(wslist, ws_left(ws));
  l_appendto(wslist, ws);
  ws_append_to_list(wslist, ws_right(ws));
}

list *ws_to_list(weight_set *ws)
{
  list *wslist = l_new();
  ws_append_to_list(wslist, ws);
  return wslist;
}


list *generate_sums(vector *a, affine *afn)
{
  list *plus_roots = afn->positive_roots;
  int n = m_dim(afn->cartan);
  int nroots = l_length(plus_roots) - n;
  int i, ok;
  int *coef;
  list *res = l_new();
  
  coef = weyl_calloc(nroots, sizeof(int));
  
  ok = 1;
  while (ok)
    {
      vector *sum = v_new(l_length(plus_roots));
      for (i = 0; i < nroots; i++)
	v_entry(sum, i) = coef[i];
      for (i = 0; i < n; i++)
	v_entry(sum, i + nroots) = v_entry(a, i);
      l_appendto(res, sum);
      
      /* find naeste */
      i = 0;
      ok = 0;
      while (ok == 0 && i < nroots)
	{
	  vector *r = l_element(plus_roots, i);
	  ok = v_lesseq(r, a);
	  if (ok)
	    {
	      v_add_multiple(a, -1, r);
	      coef[i]++;
	    }
	  else
	    {
	      v_add_multiple(a, coef[i], r);
	      coef[i] = 0;
	      i++;
	    }
	}
    }
  weyl_free(coef);
  return res;
}

void free_sums(list *sums)
{
  int i;
  for (i = 0; i < l_length(sums); i++)
    v_free(l_element(sums, i));
  l_free(sums);
}


weight_set *calc_weight_orbit(vector *w, affine *afn)
{
  int n = m_dim(afn->cartan);
  weight_set *tree = ws_new(w);
  vector *w0 = v_new(n);
  int i;
  list *lst = l_new();
  l_appendto(lst, tree);
  
  for (i = 0; i < l_length(lst); i++)
    {
      int alpha;
      weight_set *ws = l_element(lst, i);
      for (alpha = 1; alpha <= n; alpha++)
	{
	  weight_set *ws0;
	  v_copy(w0, ws_weight(ws));
	  ref_weight(w0, alpha, afn);
	  ws0 = ws_insert(tree, w0);
	  if (ws0 != NULL)
	    l_appendto(lst, ws0);
	}
    }
  v_free(w0);
  l_free(lst);
  return tree;
}
